from pathlib import Path

import pandas as pd
import zntrack


class BulkDensity(zntrack.Node):
    data_file: Path | str = zntrack.deps_path()

    results: pd.DataFrame = zntrack.plots()

    def run(self):
        pass



from pathlib import Path

import matplotlib.pyplot as plt
import MDAnalysis.analysis.msd as msd
import numpy as np
import pandas as pd
import pint
import zntrack
from MDAnalysis import Universe
from scipy.stats import linregress

from znmdakit.transformations import UnWrap, get_com_transform

ureg = pint.UnitRegistry()


class EinsteinMSD(zntrack.Node):
    universe: Universe = zntrack.deps()
    select: str = zntrack.params()
    results: pd.DataFrame = zntrack.plots(y="msd", x="lagtimes")
    apply_com_transform: bool = zntrack.params(False)

    # TODO: these two should be read from the h5md file
    timestep: float = zntrack.params(None)
    sampling_rate: int = zntrack.params(None)

    def run(self) -> None:
        universe = self.universe

        transformations = []

        if self.apply_com_transform:
            transformations.extend(get_com_transform(universe))
        universe.trajectory.add_transformations(
            *transformations, UnWrap()
        )  # UnWrap is the last one!

        MSD = msd.EinsteinMSD(universe, select=self.select, msd_type="xyz", fft=True)
        MSD.run(verbose=True)

        if self.timestep is not None and self.sampling_rate is not None:
            dt = self.timestep * self.sampling_rate
        else:
            dt = universe.trajectory.dt

        lagtimes = np.arange(MSD.n_frames) * dt

        self.results = pd.DataFrame(
            {
                "lagtimes": lagtimes,
                "msd": MSD.results.timeseries,
            }
        )
        # set the index to lagtimes
        self.results.set_index("lagtimes", inplace=True)


class SelfDiffusionFromMSD(zntrack.Node):
    """Compute the self-diffusion coefficient from the mean square displacement."""

    data: EinsteinMSD = zntrack.deps()

    start_time: float = zntrack.params(0)
    end_time: float = zntrack.params(-1)

    fit_figure: Path = zntrack.plots_path(zntrack.nwd / "msd_fit.png")

    metrics: dict = zntrack.metrics()

    def run(self):
        if self.data.timestep is None or self.data.sampling_rate is None:
            raise ValueError("timestep/sampling_rate must be set")

        # How is the msd one shorter than the lagtimes?
        linear_model = linregress(
            self.data.results.index[self.start_time : self.end_time],
            self.data.results.msd[self.start_time + 1 : self.end_time],
        )

        diff = (linear_model.slope / 6) * ureg.angstrom**2 / ureg.picosecond
        # diff = diff.to(ureg.meter**2 / ureg.second)
        diff = diff.to(ureg.angstrom**2 / ureg.nanosecond)
        # convert to nm / s

        self.metrics = {
            "slope": linear_model.slope,
            "intercept": linear_model.intercept,
            "rvalue": linear_model.rvalue,
            "diffusion": diff.magnitude,
        }

        self.fit_figure.parent.mkdir(parents=True, exist_ok=True)

        fig, ax = plt.subplots(
            figsize=(6, 4), dpi=150
        )  # Adjust figure size and resolution

        # Plot MSD and fit line with improved styling
        ax.plot(
            self.data.results.index,
            self.data.results.msd,
            label="MSD",
            lw=2,
            color="royalblue",
        )
        ax.plot(
            self.data.results.index,
            linear_model.slope * self.data.results.index + linear_model.intercept,
            label="Fit",
            lw=2,
            linestyle="--",
            color="darkorange",
        )

        # Labels with larger fonts
        ax.set_xlabel("Time t / ps", fontsize=12)
        ax.set_ylabel("MSD / Å²", fontsize=12)

        # Add vertical lines with transparency
        ax.axvline(self.start_time, color="crimson", linestyle="--", lw=1.5, alpha=0.75)
        ax.axvline(self.end_time, color="crimson", linestyle="--", lw=1.5, alpha=0.75)

        # Add a grid for readability
        ax.grid(True, linestyle="--", linewidth=0.5, alpha=0.7)

        # Add diffusion coefficient and R² value in a less intrusive way
        ax.text(
            0.05,  # Move slightly left
            0.85,  # Move slightly up
            f"Self-diffusion: {diff.magnitude:.2f} Å²/ns\nR²: {linear_model.rvalue:.2f}",
            transform=ax.transAxes,
            fontsize=10,
            bbox=dict(
                facecolor="white", alpha=0.6, edgecolor="gray", boxstyle="round,pad=0.3"
            ),
        )

        fig.savefig(self.fit_figure, bbox_inches="tight")



import znmdakit


def test_msd(universe):
    msd = znmdakit.EinsteinMSD(
        universe=universe, select="name H", timestep=0.1, sampling_rate=10
    )
    msd.run()
    assert len(msd.results) == len(universe.trajectory)



import os
import pathlib
import shutil

import dvc.cli
import git
import h5py
import MDAnalysis as mda
import pytest
from ase.data import chemical_symbols
from MDAnalysis.coordinates.H5MD import H5MDReader

FILE = pathlib.Path(__file__).parent / "lammps_npt.h5"


def get_mda_universe():
    u = mda.Universe.empty(1000, trajectory=True)
    reader = H5MDReader(FILE, convert_units=False)
    u.trajectory = reader.trajectory

    with h5py.File(FILE) as f:
        # read the types
        types = f["particles/all/species/value"][0]

    # set atoms types
    u.add_TopologyAttr("types", types)
    u.add_TopologyAttr("names", [chemical_symbols[int(t)] for t in types])

    return u


def get_znmda_universe():
    import znmdakit

    return znmdakit.Universe(FILE).get_universe()


@pytest.fixture(params=[get_mda_universe, get_znmda_universe])
def universe(request):
    return request.param()


@pytest.fixture
def repo_path(tmp_path, request):
    shutil.copy(request.module.__file__, tmp_path)
    os.chdir(tmp_path)
    git.Repo.init()
    dvc.cli.main(["init"])

    return tmp_path


@pytest.fixture
def lammps_npt():
    return FILE



from pathlib import Path

import ase
import ase.atoms
import MDAnalysis as mda
import znh5md
from ase.neighborlist import NeighborList, natural_cutoffs
from MDAnalysis.coordinates.H5MD import H5MDReader


def get_bonds(atoms: ase.Atoms) -> list[tuple[int, int]]:
    """Calculate the bonds in an ASE Atoms object.

    Parameters
    ----------
    atoms : ase.Atoms
        The ASE Atoms object for which to calculate bonds.

    Returns
    -------
    list of tuple of int
        A list of tuples, each containing two indices of bonded atoms.
    """
    # Step 1: Calculate natural cutoffs
    cutoffs = natural_cutoffs(atoms)

    # Step 2: Create a neighbor list
    nl = NeighborList(cutoffs, self_interaction=False, bothways=True)
    nl.update(atoms)

    # Step 3: Generate the bonds list
    bonds = []
    for i in range(len(atoms)):
        indices, offsets = nl.get_neighbors(i)
        for j in indices:
            if (j, i) not in bonds:  # Avoid duplicate bonds
                bonds.append((i, j))

    return bonds


def get_universe(file: Path) -> mda.Universe:
    """Create a MDAnalysis Universe from a H5MD file.

    Parameters
    ----------
    file : Path
        The path to the H5MD file.

    Returns
    -------
    mda.Universe
        The MDAnalysis Universe object.
    """

    if file.suffix not in [".h5", ".h5md", ".hdf5"]:
        raise ValueError("Currently, only HDF5 files are supported.")

    atoms: ase.Atoms = znh5md.read(file)
    # consider https://docs.mdanalysis.org/stable/documentation_pages/guesser_modules/default_guesser.html#MDAnalysis.guesser.default_guesser.DefaultGuesser.guess_bonds
    bonds = get_bonds(atoms)

    universe = mda.Universe.empty(n_atoms=len(atoms), trajectory=True)
    # https://userguide.mdanalysis.org/stable/examples/constructing_universe.html#Adding-topology-attributes
    universe.add_TopologyAttr("names", atoms.get_chemical_symbols())
    universe.add_TopologyAttr("type", atoms.get_chemical_symbols())
    universe.add_TopologyAttr("masses", atoms.get_masses())
    universe.add_TopologyAttr("bonds", bonds)

    # TODO: warning if the box changes in time! We do not support NPT yet!
    universe.dimensions = atoms.cell.cellpar()

    # add new unit translation
    # TODO: we assume these units, might not fit!
    H5MDReader._unit_translation["force"].update({"eV/Angstrom": "kJ/(mol*Angstrom)"})

    reader = H5MDReader(file, convert_units=False)
    universe.trajectory = reader

    return universe



from pathlib import Path

import zntrack
from MDAnalysis import Universe
from tqdm import tqdm

from znmdakit.transformations import UnWrap, get_com_transform


class WriteXYZ(zntrack.Node):
    universe: Universe = zntrack.deps()
    frames_path: Path = zntrack.outs_path(zntrack.nwd / "frames.xyz")

    select: str = zntrack.params("all")
    apply_com_transform: bool = zntrack.params(False)
    unwrap: bool = zntrack.params(False)

    def run(self):
        self.frames_path.parent.mkdir(parents=True, exist_ok=True)
        universe = self.universe

        transformations = []
        if self.apply_com_transform:
            transformations.extend(get_com_transform(universe))
        if self.unwrap:
            transformations.append(UnWrap())
        if len(transformations) > 0:
            universe.trajectory.add_transformations(*transformations)

        atoms_group = universe.select_atoms(self.select)
        print(f"Writing {len(atoms_group)} atoms to {self.frames_path}")

        with open(self.frames_path, "w") as f:
            # TODO: write znh5md
            for idx, ts in tqdm(enumerate(universe.trajectory)):
                f.write(f"{len(atoms_group)}\n")
                f.write(f"Frame {ts.frame}\n")
                # TODO: somehow define atom type per residue for ASE?
                for atom in atoms_group:
                    f.write(
                        f"X {atom.position[0]} {atom.position[1]} {atom.position[2]}\n"
                    )



import numpy as np
from MDAnalysis import AtomGroup
from MDAnalysis.transformations import TransformationBase
from tqdm import tqdm


class UnWrap(TransformationBase):
    prev = None
    # def __init__(self, **kwargs):
    #     super().__init__(parallelizable=False, **kwargs)
    #     self.prev = None

    def _transform(self, ts):
        if ts.frame == 0:
            self.prev = ts.positions.copy()
            return ts

        assert np.all(ts.dimensions[3:] == [90, 90, 90])

        while np.any(ts.positions - self.prev > 0.5 * ts.dimensions[:3]):
            ts.positions = np.where(
                ts.positions - self.prev > 0.5 * ts.dimensions[:3],
                ts.positions - ts.dimensions[:3],
                ts.positions,
            )

        while np.any(ts.positions - self.prev < -0.5 * ts.dimensions[:3]):
            ts.positions = np.where(
                ts.positions - self.prev < -0.5 * ts.dimensions[:3],
                ts.positions + ts.dimensions[:3],
                ts.positions,
            )
        # assert that there is no jump larger than half the box size
        assert np.all(np.abs(ts.positions - self.prev) < 0.5 * ts.dimensions[:3])

        self.prev = ts.positions.copy()
        return ts


# https://gist.github.com/orbeckst/b611626e7640f399580e233195c6429c
# Center of Mass Transformation

# TODO: consider using the "BeadGroup" suggested in the gist instead


class COMTransform(TransformationBase):
    def __init__(self, reference: AtomGroup, name: str):
        self.reference = reference
        self.com_atoms = reference.select_atoms(f"name {name}")

        # sanity check
        a = self.get_com().shape
        b = self.com_atoms.positions.shape
        if a != b:
            raise ValueError(f"Shape mismatch: {a} != {b}")

    def get_com(self):
        return self.reference.center_of_mass(unwrap=True, compound="fragments")

    def __call__(self, ts):
        self.com_atoms.positions = self.get_com()
        return ts


def get_com_transform(universe) -> list:
    # TODO: sanity check, write to XYZ, construct a base Node and one node that just writes the transformed trajectory to disk!
    # when writing to disk, allow for a list of selection strings
    transformations = {}
    for residue in tqdm(universe.residues, desc="Updating atom names"):
        if len(residue.atoms) > 1:
            # we will later update the positions of the first atom in each residue
            residue.atoms[0].name = "COM"
    for residue in tqdm(universe.residues, desc="Preparing COMTransform"):
        if len(residue.atoms) > 1:
            if residue.resname not in transformations:
                transformations[residue.resname] = COMTransform(
                    reference=universe.select_atoms(f"resname {residue.resname}"),
                    name="COM",
                )
    return list(transformations.values())



from znmdakit.nodes.ionic_conductivity import NernstEinsteinIonicConductivity
from znmdakit.nodes.mean_square_displacement import EinsteinMSD, SelfDiffusionFromMSD
from znmdakit.nodes.radial_distribution_function import InterRDF
from znmdakit.nodes.universe import Universe
from znmdakit.nodes.write_xyz import WriteXYZ

__all__ = [
    "InterRDF",
    "EinsteinMSD",
    "Universe",
    "WriteXYZ",
    "SelfDiffusionFromMSD",
    "NernstEinsteinIonicConductivity",
]



from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import zntrack
from MDAnalysis import Universe
from MDAnalysis.analysis import rdf

from znmdakit.transformations import get_com_transform


class InterRDF(zntrack.Node):
    universe: Universe = zntrack.deps()
    g1: str = zntrack.params()
    g2: str = zntrack.params()
    nbins: int = zntrack.params()
    range: str | tuple = zntrack.params("auto")

    apply_com_transform: bool = zntrack.params(
        False
    )  # replace the position of the first atom in each residue with the center of mass and name it COM

    figure_path: Path = zntrack.outs_path(zntrack.nwd / "rdf.png")

    results: pd.DataFrame = zntrack.plots(y="g(r)", x="r")

    def run(self):
        universe = self.universe

        if self.apply_com_transform:
            transformations = get_com_transform(universe)
            universe.trajectory.add_transformations(*transformations)

        RDF = rdf.InterRDF(
            universe.select_atoms(self.g1),
            universe.select_atoms(self.g2),
            nbins=1000,
            range=(0, universe.dimensions[0] / 2)
            if self.range == "auto"
            else self.range,
        )
        RDF.run(verbose=True)

        self.results = pd.DataFrame(
            {"r": RDF.results.bins[1:], "g(r)": RDF.results.rdf[1:]}
        )

        self.results.set_index("r", inplace=True)
        self.figure_path.parent.mkdir(parents=True, exist_ok=True)
        figure = self.get_figure()
        figure.savefig(self.figure_path, bbox_inches="tight")

    def get_figure(self):
        sns.set_theme(style="whitegrid")  # Use a clean seaborn style

        fig, ax = plt.subplots(
            figsize=(6, 4), dpi=100
        )  # Set figure size and resolution
        ax.plot(self.results.index, self.results["g(r)"], color="tab:blue", lw=2)

        ax.set_ylabel("Radial Distribution Function, g(r)", fontsize=12)
        ax.set_xlabel("Distance, r (Å)", fontsize=12)
        ax.set_title(f"InterRDF: {self.g1} - {self.g2}", fontsize=14, fontweight="bold")

        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)

        return fig



import znmdakit


def test_workflow(repo_path, lammps_npt):
    project = znmdakit.Project()

    with project:
        u = znmdakit.Universe(lammps_npt)
        rdf = znmdakit.RDF(universe=u, select="name H", nbins=100)
        msd = znmdakit.EinsteinMSD(
            universe=u, select="name H", timestep=0.1, sampling_rate=10
        )

    project.run()

    rdf.load()
    msd.load()

    print(rdf.results)
    print(msd.results)



import logging
from pathlib import Path

import ase
import znh5md
import zntrack
from rdkit2ase import smiles2atoms

from znmdakit.utils import get_universe

log = logging.getLogger(__name__)

# class OverwriteDict(t.TypedDict):
#     timestep: float
#     sampling_rate: int


class Universe(zntrack.Node):
    data_file: str = zntrack.deps_path(None)
    data: list[ase.Atoms] = zntrack.deps(None)
    residues: dict[str, str] = zntrack.params()  # dict[id, smiles]
    # overwrite: OverwriteDict = zntrack.params(default_factory=dict)

    frames_path: Path = zntrack.outs_path(zntrack.nwd / "frames.h5")

    def run(self):
        if self.data_file is not None and self.data is not None:
            raise ValueError("Either data_file or data must be provided")

        self.frames_path.parent.mkdir(parents=True, exist_ok=True)
        io = znh5md.IO(
            self.frames_path, store="time", save_units=False
        )  # ensure H5Reader can read it.

        if self.data is not None:
            io.extend(self.data)
        else:
            io.append(ase.Atoms("H", positions=[[0, 0, 0]]))

    @property  # cached property needed?
    def universe(self):
        log.critical("Constructing universe")
        if self.data_file is not None:
            universe = get_universe(Path(self.data_file))
        else:
            universe = get_universe(self.frames_path)

        residues = {k: smiles2atoms(smiles=v) for k, v in self.residues.items()}

        universe.add_TopologyAttr("resnames")

        if len(self.residues) > 0:
            for mol in universe.atoms.fragments:
                for residue in residues:
                    # We assign the residues by matching the chemical symbols of the atoms
                    # There is no graph comparison to ensure the atoms are connected in the same way!
                    # If the connectivity graph changes over time, this is not captured and thus not suitable
                    # for reactive systems!
                    if sorted(mol.names) == sorted(
                        residues[residue].get_chemical_symbols()
                    ):
                        # https://docs.mdanalysis.org/stable/documentation_pages/core/universe.html#MDAnalysis.core.universe.Universe.add_Residue
                        mda_residue = universe.add_Residue(
                            resid=len(universe.residues),
                            resname=residue,
                            # resnum=len(mda_residues),
                        )
                        universe.atoms[mol.indices].residues = mda_residue
                        break
                else:
                    log.warning(
                        f"Could not find residue for molecule {sorted(mol.names)} in residues {list(sorted(x.get_chemical_symbols()) for x in residues.values())}"
                    )

        return universe



import typing as t

import pint
import zntrack
from MDAnalysis import Universe

ureg = pint.UnitRegistry()


class NEIons(t.TypedDict):
    diffusion: float
    charge: int
    select: str


class NEParams(t.TypedDict):
    temperature: float
    ions: t.Dict[str, NEIons]


class NernstEinsteinIonicConductivity(zntrack.Node):
    """Compute the ionic conductivity using the Nernst-Einstein equation.


    References
    ----------
    https://www.sciencedirect.com/science/article/pii/S2772422024000120
    https://jcheminf.biomedcentral.com/articles/10.1186/s13321-023-00687-y
    https://journals.aps.org/prl/pdf/10.1103/PhysRevLett.122.136001

    """

    universe: Universe = zntrack.deps()
    params: NEParams = zntrack.params()

    metrics: dict = zntrack.metrics()

    def run(self):
        volume = (
            self.universe.dimensions[0]
            * self.universe.dimensions[1]
            * self.universe.dimensions[2]
        ) * ureg.angstrom**3

        # e^2 / (T * kB * V)
        prefactor = (ureg.elementary_charge**2) / (
            self.params["temperature"] * ureg.kelvin * ureg.boltzmann_constant * volume
        )
        # is it 3 * volume or just volume?

        diff_x_charge = []
        for ion_data in self.params["ions"].values():
            atoms_group = self.universe.select_atoms(ion_data["select"])

            # D * z^2 per ion (thus * n_ions)
            n_ions = len(atoms_group.fragments)
            diff = ion_data["diffusion"] * ureg.angstrom**2 / ureg.nanosecond
            value = diff * ion_data["charge"] ** 2 * n_ions

            diff_x_charge.append(value)

        sigma_NE = prefactor * sum(diff_x_charge)
        sigma_NE = sigma_NE.to("S/m")

        self.metrics = {"Nernst-Einstein ionic conductivity": sigma_NE.magnitude}
